@@ -1,6 +1,5 @@
 use clap::ArgMatches;
 use clap_complete::Shell::*;
-use rayon::prelude::*;
 use regex::Regex;
 use rusqlite::Error;
 
@@ -13,31 +12,6 @@ use crate::Zettel;
 use crate::cli;
 use crate::io::{abs_path, file_exists};
 
-pub fn sync(matches: &ArgMatches, cfg: &ConfigOptions) -> Result<(), Error>
-{
-    let project = realproject(if let Some(p) = matches.get_one::<String>("PROJECT") {
-                                  p
-                              } else {
-                                  ""
-                              });
-    if let Some(title) = matches.get_one::<String>("CREATE") {
-        create(cfg, title, project)?;
-    } else if let Some(path) = matches.get_one::<String>("UPDATE") {
-        update(cfg, path)?;
-    } else if let Some(title) = matches.get_one::<String>("MOVE") {
-        mv(cfg, title, project)?;
-    } else if matches.contains_id("RENAME") {
-        let args_arr = matches.get_many::<String>("RENAME")
-                              .unwrap_or_default()
-                              .map(|a| a.to_string())
-                              .collect::<Vec<_>>();
-        rename(cfg, &args_arr)?;
-    } else if matches.get_flag("GENERATE") {
-        generate(cfg)?;
-    }
-    Ok(())
-}
-
 /// A printer that prints. Because it's convenient.
 struct Printer
 {
@@ -62,11 +36,6 @@ impl Printer
         self.zettel = new_zettel;
     }
 
-    fn set_single_zettel(&mut self, new_zettel: Zettel)
-    {
-        self.zettel = vec![new_zettel];
-    }
-
     fn set_additional(&mut self, new_additional: Vec<String>)
     {
         self.additional = new_additional;
@@ -82,6 +51,12 @@ impl Printer
         self.link_separator = new_separator;
     }
 
+    fn print_one(&mut self, cfg: &ConfigOptions, zettel: Zettel)
+    {
+        self.zettel = vec![zettel];
+        self.print(cfg);
+    }
+
     /// Abracadabra, yadda yadda. Print everything properly.
     fn print(&mut self, cfg: &ConfigOptions)
     {
@@ -100,18 +75,7 @@ impl Printer
             result = result.replace("%P", &z.filename(cfg));
             result = result.replace("%l", &z.links.join(&self.link_separator));
             result = result.replace("%a", a);
-            // Based on the provided ConfigOptions, we may or may not get the backlinks for the given
-            // Zettel, so if we don't, we just consume the `%b` token and move on
-            if result.contains("%b") {
-                let maybe_get_backlinks = || -> Result<Vec<String>, Error> {
-                    let all = Database::new(&cfg.db_file())?.all()?;
-                    let bks = backlinks(&all, &z.title, true);
-                    Ok(bks.iter().map(|z| z.title.clone()).collect())
-                };
-                if let Ok(bks) = maybe_get_backlinks() {
-                    result = result.replace("%b", &bks.join(&self.link_separator));
-                }
-            }
+            result = result.replace("%b", &z.backlinks.join(&self.link_separator));
 
             println!("{}", result);
         }
@@ -129,10 +93,38 @@ impl Default for Printer
     }
 }
 
+pub fn sync(matches: &ArgMatches, cfg: &ConfigOptions) -> Result<(), Error>
+{
+    Database::new(&cfg.db_file())?.init()?;
+
+    let project = realproject(if let Some(p) = matches.get_one::<String>("PROJECT") {
+                                  p
+                              } else {
+                                  ""
+                              });
+    if let Some(title) = matches.get_one::<String>("CREATE") {
+        create(cfg, title, project)?;
+    } else if let Some(path) = matches.get_one::<String>("UPDATE") {
+        update(cfg, path)?;
+    } else if let Some(title) = matches.get_one::<String>("MOVE") {
+        mv(cfg, title, project)?;
+    } else if matches.contains_id("RENAME") {
+        let args_arr = matches.get_many::<String>("RENAME")
+                              .unwrap_or_default()
+                              .map(|a| a.to_string())
+                              .collect::<Vec<_>>();
+        rename(cfg, &args_arr)?;
+    } else if matches.get_flag("GENERATE") {
+        generate(cfg)?;
+    }
+    Ok(())
+}
+
 /// Query the database, applying various filters if proivded
 pub fn query(matches: &ArgMatches, cfg: &ConfigOptions) -> Result<(), Error>
 {
     let db = Database::new(&cfg.db_file())?;
+    db.init()?;
 
     let mut zs: Vec<Zettel> = db.all()?;
     let mut printer = Printer::default();
@@ -145,17 +137,6 @@ pub fn query(matches: &ArgMatches, cfg: &ConfigOptions) -> Result<(), Error>
     if let Some(project) = matches.get_one::<String>("PROJECT") {
         zs = filter_project(zs, realproject(project), exact);
     }
-    if let Some(text) = matches.get_one::<String>("TEXT_REGEX") {
-        let vs = filter_text(zs.clone(), text, cfg);
-        let mut texts: Vec<String> = vec![];
-        zs = vec![]; // reset Zettel vector and append indiivdually
-                     // maybe speed this up by using unzip? I digress
-        for (z, t) in vs {
-            zs.push(z);
-            texts.push(t);
-        }
-        printer.set_additional(texts);
-    }
     if let Some(tag) = matches.get_one::<String>("TAG") {
         zs = filter_tag(zs, tag, exact);
     }
@@ -165,6 +146,19 @@ pub fn query(matches: &ArgMatches, cfg: &ConfigOptions) -> Result<(), Error>
     if let Some(links_to) = matches.get_one::<String>("BACKLINKS") {
         zs = intersect(&zs, &backlinks(&zs, links_to, exact));
     }
+    if let Some(text) = matches.get_one::<String>("TEXT_REGEX") {
+        let vs = filter_text(zs.clone(), text, cfg);
+        let mut texts: Vec<String> = vec![];
+        let mut found = vec![];
+        // maybe speed this up by using unzip? I digress
+        for (z, t) in vs {
+            found.push(z);
+            texts.push(t);
+        }
+        printer.set_additional(texts);
+        zs = intersect(&zs, &found);
+    }
+
     if matches.get_flag("LONERS") {
         zs = filter_isolated(zs);
     }
@@ -196,6 +190,7 @@ fn replace_literals(s: &str) -> String
 pub fn ls(matches: &ArgMatches, cfg: &ConfigOptions) -> Result<(), Error>
 {
     let db = Database::new(&cfg.db_file())?;
+    db.init()?;
 
     let obj = if let Some(m) = matches.get_one::<String>("OBJECT") {
         m
@@ -218,13 +213,9 @@ pub fn ls(matches: &ArgMatches, cfg: &ConfigOptions) -> Result<(), Error>
 /// Generate completions for a shell
 pub fn compl(matches: &ArgMatches) -> Result<(), Error>
 {
-    let shell = if let Some(m) = matches.get_one::<String>("OBJECT") {
-        m
-    } else {
-        ""
-    };
+    let shell = matches.get_one::<String>("SHELL").unwrap();
 
-    let sh = match shell {
+    let sh = match shell.as_str() {
         "zsh" => Some(Zsh),
         "bash" => Some(Bash),
         "fish" => Some(Fish),
@@ -305,9 +296,8 @@ fn filter_tag(zs: Vec<Zettel>, pattern: &str, exact: bool) -> Vec<Zettel>
 /// Keep only those Zettel that neither link to other notes, nor have links pointing to them
 fn filter_isolated(zs: Vec<Zettel>) -> Vec<Zettel>
 {
-    zs.clone()
-      .into_iter()
-      .filter(|z| z.links.is_empty() && backlinks(&zs, &z.title, true).is_empty())
+    zs.into_iter()
+      .filter(|z| z.links.is_empty() && z.backlinks.is_empty())
       .collect()
 }
 
@@ -376,22 +366,25 @@ fn create(cfg: &ConfigOptions, title: &str, project: &str) -> Result<(), Error>
     }
 
     let exists_in_fs = file_exists(&zettel.filename(cfg));
-    let exists_in_db = db.all()?.into_par_iter().any(|z| z == zettel);
+    let exists_in_db = db.find_by_title(&zettel.title)?.len() == 1;
 
     // If the corresponding file exists and there's an entry in the database, abort.
+    // If there's an entry in the database but no corresponding file, replace the database entry
     // If there's a file but there's no entry in the database, create an entry.
     // Otherwise, create a new file from template and add a database entry.
     if exists_in_fs && exists_in_db {
         eprintln!("error: couldn't create new Zettel: one with the same title already exists");
         return Ok(());
-    } else if exists_in_fs {
-        println!("file exists in the filesystem but not in the database; added entry");
+    } else if !exists_in_fs && exists_in_db {
+        eprintln!("Zettel exists in the database but not on the filesystem; replaced entry");
+        db.delete(&zettel)?;
+        // saved outside of the loop
+    } else if exists_in_fs && !exists_in_db {
+        eprintln!("Zettel exists on the filesystem but not in the database; added entry");
         // saved outside of the loop
     } else {
         zettel.create(cfg);
-        let mut printer = Printer::default();
-        printer.set_single_zettel(zettel.clone());
-        printer.print(cfg);
+        Printer::default().print_one(cfg, zettel.clone());
     }
     db.save(&zettel)?;
 
@@ -443,8 +436,6 @@ fn rename(cfg: &ConfigOptions, arr: &[String]) -> Result<(), Error>
         db.change_title(old_zettel, &new_title).unwrap();
         // It's not enough that we renamed the file. We need to update all references to it!
         let backlinks = backlinks(&db.all()?, &old_title, true);
-        // for some reason rustfmt has absolutely cursed formatting here. this is not my fault, I
-        // swear
         backlinks.iter().for_each(|bl| {
                             let contents = crate::io::file_to_string(&bl.filename(cfg));
                             // The link might span over multiple lines. We must account for that
@@ -491,6 +482,7 @@ fn mv(cfg: &ConfigOptions, pattern: &str, project: &str) -> Result<(), Error>
         let new_notes = zs.iter().map(|z| Zettel { title: z.title.clone(),
                                                    project: project.to_string(),
                                                    links: z.links.clone(),
+                                                   backlinks: z.backlinks.clone(),
                                                    tags: z.tags.clone() });
         let pairs = zs.iter().zip(new_notes);
         pairs.for_each(|(old, new)| {
@@ -528,9 +520,9 @@ fn generate(cfg: &ConfigOptions) -> Result<(), Error>
 {
     let start = std::time::Instant::now();
 
-    let mem_db = Database::in_memory(&cfg.db_file())?;
+    let mem_db = Database::new_in_memory(&cfg.db_file())?;
     mem_db.init()?;
-    mem_db.generate(cfg);
+    mem_db.generate(cfg)?;
     mem_db.write_to(&cfg.db_file())?;
 
     println!("database generated successfully, took {}ms",
